# SPDX-License-Identifier: GPL-3.0-only

APP_ID = "org.gnome.World.Secrets"
VERSION = "13.0.1"
GETTEXT_PACKAGE = "secrets"
LOCALEDIR = "/usr/share/locale"
NAME = "Secrets"
PKGDATADIR = "/usr/share/secrets"
PROFILE = ""
PKCS11_LIB = "/usr/lib/opensc-pkcs11.so"

IS_DEVEL = PROFILE == "Devel"
SUFFIX = " (Devel)" if IS_DEVEL else ""
SHORT_NAME = GETTEXT_PACKAGE
